// $Header: /teamproj/cvsroot/teamb/stylesrc/au/edu/swin/synergy/style/BFileParser.java,v 1.4 1999/09/03 04:54:41 tp079519 Exp $

/**
 * Synergy Java code format utility
 *
 * John Newbigin
 *
 *
 * ToDo:
 *  Work out how to work out the context of stuff.......
 *  allign = signs
 *  what makes a variable decleration
 *  what makes a method
 *  where are { mandatitory?
 *
 * @see spec.txt
 * @concurrency N/A
 * @copyright Synergy Software 1999
 * @rref N/A
 * @author John Newbigin
 */

package au.edu.swin.synergy.style;

import java.util.Vector;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.OutputStream;
import java.lang.String;

/**
 * DoTo:...
 *  Look inside the class. (indent = 1)
 *  If a line has an = then is is an instanceVariable IV
 *  If a line has a set of () (before the =) is is a method M
 *  If a line has the word class it is an inner class IC
 *  there can be comments
 *  and anything else is an error for now.
 *
 *  I still need to address global scope issues
 *  such as the class name/file name
 *  package name and imports
 *
 *  if a TT_COMMENT follows a ";", it gets moved to the next line.
 *  if you have static init array, the trailing ";" get's lost when the code it re-arranged.
 *  if you have more than one outer class in a .java file, code re-arrangement might break it.
 *
 */

public class BFileParser
{
    /**
     *  Parse the java file and make it conform to the Synergy style...
     *  This does not use a StreamTokenizer as it does not provide the required
     *  Flexability of the handling of comments.
     */
    
    public static final int kIndentLevel = 4;
    
    String thisLine = new String();
    int thisIndent = 0;
    int printCount = 0;
    // the number of printed tokens on the line
    boolean dontPrintBlankLine = false;
    int bracketIndent = 0;
    
    int missingBraceState = 0;
    int braceBracketLevel = 0;
    boolean warnMissingBrace = false;
    
    boolean staticTagsPrinted = false;
    boolean printWarning = false;
    boolean foundHeader = false;
    
    Vector codeSections = null;
    BCodeSection currentSection = null;
    private OutputStream outStream = null;
    
    private static final String kJAVADOCstart = "/**";
    private static final String kJAVADOCend = "*/";
    
    private Vector ruleSet = null;
    
    private String className = "unknown";
    
    private boolean again = false;
    // set this true if we need another parse
    
    /**
     * BFileParser
     * @post The ruleset is initalised
     * @pre none
     * @throws none
     * @returns none
     * @param none
     */
    
    public BFileParser()
    {
        createRuleSet();
        
        codeSections = new Vector();
        startSection();
    }
    
    /**
     * addMissingTags
     * @post any missing tags are added to the code section
     * @pre
     * @throws none
     * @returns void
     * @param a code section
     */
    
    void addMissingTags( BCodeSection section )
    {
        BStringList required = new BStringList();
        
        if( section.getType() == BCodeSection.STATIC )
        {
            if( staticTagsPrinted == false )
            {
                required.add( "author" );
                required.add( "rref" );
                required.add( "copyright" );
                required.add( "concurrency" );
                required.add( "see" );
                staticTagsPrinted = true;
            }
        }
        else if( section.getType() == BCodeSection.METHOD )
        {
            required.add( "param" );
            required.add( "return" );
            required.add( "throws" );
            required.add( "pre" );
            required.add( "post" );
            
        }
        else if( section.getType() == BCodeSection.IV )
        {
            
        }
        else if( section.getType() == BCodeSection.CLASS )
        {
            
        }
        
        BStringList tags = section.getTagList();
        for( int i = 0; i < tags.getCount(); i++ )
        {
            // see if the tag is in the required list
            for( int j = 0; j < required.getCount(); j++ )
            {
                if( required.get( j ).equalsIgnoreCase( tags.get( i ) ) )
                {
                    required.remove( j );
                    break;
                }
            }
        }
        if( required.getCount() > 0 )
        {
            // put in the new tags....
            int insertBefore = - 1;
            
            // we need to put them into the existing javadoc if possible.
            for( int i = 0; i < section.getCount(); i++ )
            {
                String line = section.getLine( i );
                if( line.trim().equals( "*/" ) )
                {
                    // we found the javadoc...
                    insertBefore = i;
                    break;
                }
            }
            
            BStringList lines = new BStringList();
            
            if( section.getName() == null )
            {
                lines.add( " * Please complete these missing tags" );
            }
            else
            {
                lines.add( " * Please complete the missing tags for " + section.getName() );
            }
            for( int i = 0; i < required.getCount(); i++ )
            {
                lines.add( " * @" + required.get( i ) );
            }
            
            if( insertBefore == - 1 )
            {
                // we'll just stick it at the start
                section.prependLine( " */" );
                for( int i = lines.getCount(); i > 0; i-- )
                {
                    section.prependLine( lines.get( i - 1 ) );
                }
                
                section.prependLine( "/**" );
            }
            else
            {
                for( int i = 0; i < lines.getCount(); i++ )
                {
                    section.insert( insertBefore + i, lines.get( i ) );
                }
            }
            printWarning = true;
        }
        
        // we could consolidate all javadoc into one block.
    }
    
    /**
     * startSection
     * @post a new code section is started
     * @pre there is a current code section
     * @throws none
     * @returns void
     * @param void
     */
    
    void startSection()
    {
        currentSection = new BCodeSection();
        codeSections.add( currentSection );
    }
    
    
    /**
     * println
     * @post a new line is added to the code section
     * @pre
     * @throws none
     * @returns void
     * @param void
     */
    
    void println()
    {
        if( dontPrintBlankLine && ( printCount == 0 ) )
        {
            // we will skip this blank line
            dontPrintBlankLine = false;
        }
        else
        {
            String str = new String();
            for( int i = 0; i < ( thisIndent * kIndentLevel ); i++ )
            {
                str = str + " ";
            }
            str = str + thisLine;
            currentSection.prinln( str );
            thisLine = new String();
            printCount = 0;
        }
    }
    
    
    /**
     * print
     * @post the string is added to the current line
     * @pre the current line exists
     * @throws none
     * @returns void
     * @param string to add
     */
    
    void print( String str )
    {
        thisLine = thisLine + str;
        printCount++;
    }
    
    /**
     * parse
     * @post
     * @pre
     * @throws
     * @returns
     * @param
     */
    
    boolean parse( File file, OutputStream outputStream )
    {
        if( outputStream == null )
        {
            outStream = System.out;
        }
        else
        {
            outStream = outputStream;
        }
        try
        {
            BTokeniser tokeniser = new BTokeniser( file );
            
            int indent = 0;
            boolean finished = false;
            BToken previous = null;
            BToken token = null;
            
            int classLevelState = 0;
            int fileLevelState = 0;
            
            String nameWouldBe = null;
            // the name that a method, class or variable would have, assuming that there is a method, variable or class....
            while( ! finished )
            {
                token = tokeniser.getToken();
                
                // look at the context, before we process for looks....
                if( indent == 0 )
                {
                    // find the class name
                    // find the 
                    // the states are 
                    // 0 - ?
                    // 1 - looking
                    // 2 - import
                    // 3 - class
                    if( token.getToken() == BTokeniser.TT_COMMENT )
                    {
                        if( token.getString().indexOf( "$Header" ) >= 0 )
                        {
                            foundHeader = true;
                        }
                    }
                    if( fileLevelState == 2 )
                    {
                        // read the import till a ;
                    }
                    else if( fileLevelState == 3 )
                    {
                        // read the class name
                        className = token.getString();
                        //System.out.println("cass name is " + className);
                        fileLevelState = 1;
                    }
                    else if( fileLevelState == 4 )
                    {
                        
                    }
                    else
                    {
                        if( token.getToken() == BTokeniser.TT_WORD )
                        {
                            if( token.getString().equals( "class" ) )
                            {
                                fileLevelState = 3;
                            }
                        }
                    }
                }
                if( indent == 1 )
                {
                    // this is a class level token...
                    // states are 
                    // 0 - start of section
                    // 1 - looking
                    // 2 - is an iv
                    // 3 - could be an inner class
                    
                    if( classLevelState == 0 )
                    {
                        startSection();
                        classLevelState = 1;
                    }
                    
                    if( token.getToken() == BTokeniser.TT_SYMBOL )
                    {
                        if( token.getString().equals( ";" ) )
                        {
                            // end of an i1 line
                            if( classLevelState == 1 )
                            {
                                currentSection.setName( nameWouldBe );
                            }
                            classLevelState = 0;
                        }
                        else if( token.getString().equals( "=" ) )
                        {
                            // and = means an iv, ignore to eol
                            currentSection.setName( nameWouldBe );
                            classLevelState = 2;
                            currentSection.setType( BCodeSection.IV );
                        }
                        else if( token.getString().equals( "{" ) )
                        {
                            // a method.  Ignore to }
                            if( classLevelState == 3 )
                            {
                                currentSection.setName( nameWouldBe );
                                currentSection.setType( BCodeSection.CLASS );
                            }
                            else if( classLevelState == 2 )// already an IV
                            {
                                // do nothing
                                // flag that we need a trailing ";"
                            }
                            else// must be a method
                            {
                                currentSection.setType( BCodeSection.METHOD );
                            }
                            classLevelState = 0;
                        }
                        else if( token.getString().equals( "(" ) )
                        {
                            currentSection.setName( nameWouldBe );
                        }
                    }
                    else if( token.getToken() == BTokeniser.TT_WORD )
                    {
                        if( token.getString().equals( "class" ) )
                        {
                            // class keyword found
                            //print(" /* could be an inner class */ " );
                            classLevelState = 3;
                        }
                        else if( token.getString().equals( "private" ) )
                        {
                            currentSection.setScope( BCodeSection.PRIVATE );
                        }
                        else if( token.getString().equals( "protected" ) )
                        {
                            currentSection.setScope( BCodeSection.PROTECTED );
                        }
                        else if( token.getString().equals( "public" ) )
                        {
                            currentSection.setScope( BCodeSection.PUBLIC );
                        }
                        else
                        {
                            nameWouldBe = token.getString();
                        }
                    }
                }
                
                // look for missing braces...
                if( missingBraceState == 0 )
                {
                    if( token.getToken() == BTokeniser.TT_WORD )
                    {
                        if( token.getString().equals( "for" ) || token.getString().equals( "while" ) )
                        {
                            missingBraceState = 1;
                            braceBracketLevel = 0;
                            //							braceBracketCount = 0;
                        }
                    }
                }
                else if( missingBraceState == 1 )
                {
                    if( token.getToken() == BTokeniser.TT_SYMBOL )
                    {
                        if( token.getString().equals( "(" ) )
                        {
                            braceBracketLevel++;
                        }
                        else if( token.getString().equals( ")" ) )
                        {
                            braceBracketLevel--;
                            if( braceBracketLevel == 0 )
                            {
                                missingBraceState = 2;
                            }
                        }
                    }
                }
                else if( missingBraceState == 2 )
                {
                    if( token.getToken() == BTokeniser.TT_SYMBOL )
                    {
                        if( token.getString().equals( "{" ) )
                        {
                            missingBraceState = 0;
                            // back to OK.
                        }
                        else if( token.getString().equals( ";" ) )
                        {
                            missingBraceState = 0;
                            // back to OK.
                        }
                        else
                        {
                            // we need to insert a set of braces!
                            print( "{" );
                            missingBraceState = 3;
                        }
                    }
                    else if( token.getToken() == BTokeniser.TT_NEWLINE )
                    {
                        // skip
                    }
                    else if( token.getToken() == BTokeniser.TT_COMMENT )
                    {
                        // skip
                    }
                    else if( token.getToken() == BTokeniser.TT_BLOCKCOMMENT )
                    {
                        // skip
                    }
                    else
                    {
                        // we need to insert a set of braces!
                        print( "{" );
                        missingBraceState = 3;
                    }
                }
                else if( missingBraceState == 3 )
                {
                    // check the bracket level (of '(' and ')')...
                    if( token.getToken() == BTokeniser.TT_SYMBOL )
                    {
                        if( token.getString().equals( ";" ) && ( braceBracketLevel == 0 ) )
                        {
                            // how do we add the '}'?
                            token.setString( ";}" );
                            missingBraceState = 0;
                            warnMissingBrace = true;
                        }
                        else if( token.getString().equals( "(" ) )
                        {
                            braceBracketLevel++;
                        }
                        else if( token.getString().equals( ")" ) )
                        {
                            braceBracketLevel--;
                        }
                    }
                }
                
                switch( token.getToken() )
                {
                    case BTokeniser.TT_EOF:
                    {
                        finished = true;
                        
                        // we can now check a whole lot of things...
                        // such as non-existant required sections...
                        break;
                    }
                    
                    case BTokeniser.TT_NEWLINE:
                    {
                        println();
                        thisIndent = indent;
                        break;
                    }
                    
                    case BTokeniser.TT_WORD:
                    case BTokeniser.TT_STRING:
                    case BTokeniser.TT_CHAR:
                    case BTokeniser.TT_NUMBER:
                    {
                        if( useWhitespace( token, previous ) )
                        {
                            print( " " );
                        }
                        print( token.getString() );
                        break;
                    }
                    
                    case BTokeniser.TT_COMMENT:
                    {
                        // special new line skip....?
                        boolean skipNewLine = false;
                        if( previous != null )
                        {
                            if( previous.getToken() != BTokeniser.TT_NEWLINE )
                            {
                                skipNewLine = true;
                            }
                        }
                        
                        if( skipNewLine )
                        {
                            // if the end of the previous line as a ";" then, we need to undo the new line! (Arrgh)
                        }
                        
                        
                        print( token.getString() );
                        break;
                    }
                    ;
                    
                    case BTokeniser.TT_BLOCKCOMMENT:
                    {
                        // see if this is a comment, or a javadoc....
                        if( token.getString().startsWith( kJAVADOCstart ) && ( token.getString().length() > 4 ) )
                        {
                            BStringList stringList = new BStringList();
                            stringList.assign( token.getString() );
                            
                            // make sure that there is no text on the fist line
                            if( stringList.getCount() > 0 )
                            {
                                String start = stringList.get( 0 ).trim();
                                if( start.equals( kJAVADOCstart ) == false )
                                {
                                    // we have to spilt up this line
                                    stringList.remove( 0 );
                                    String newString = start.substring( 3 );
                                    stringList.insert( 0, newString );
                                    stringList.insert( 0, kJAVADOCstart );
                                }
                            }
                            
                            // make sure that there is no text on the last line
                            if( stringList.getCount() > 1 )
                            {
                                String last = stringList.get( stringList.getCount() - 1 ).trim();
                                if( last.equals( kJAVADOCend ) == false )
                                {
                                    // we have to split up the last line
                                    stringList.remove( stringList.getCount() - 1 );
                                    String newString = last.substring( 0, last.length() - 2 );
                                    stringList.add( newString );
                                    stringList.add( kJAVADOCend );
                                }
                            }
                            
                            for( int i = 0; i < stringList.getCount(); i++ )
                            {
                                String line = stringList.get( i ).trim();
                                if( i == 0 )
                                {
                                    // the first line is OK.
                                }
                                else if( i == stringList.getCount() - 1 )
                                {
                                    // the last line is OK.
                                    line = " " + line;
                                    // do the 1 space indent
                                }
                                else
                                {
                                    // see if the line starts with a *
                                    if( line.startsWith( "*" ) == false )
                                    {
                                        line = "* " + line;
                                    }
                                    line = " " + line;
                                    // do the 1 space indent
                                    // look for javadoc tags
                                    if( line.indexOf( '@' ) >= 0 )
                                    {
                                        // extract the tag name...
                                        int start = line.indexOf( '@' );
                                        int end = line.indexOf( ' ', start );
                                        if( end <= start )
                                        {
                                            end = line.length();
                                        }
                                        String tag = line.substring( start + 1, end );
                                        //System.out.println("found @" + tag);
                                        currentSection.addTag( tag.trim() );
                                    }
                                }
                                print( line );
                                println();
                                dontPrintBlankLine = true;
                            }
                        }
                        else
                        {
                            print( token.getString() );
                        }
                        break;
                    }
                    ;
                    
                    
                    case BTokeniser.TT_SYMBOL:
                    {
                        if( token.getString().equals( "{" ) )
                        {
                            // are there any tokens on this line already???
                            if( printCount > 0 )
                            {
                                // we need a new line...
                                println();
                            }
                            indent++;
                            print( token.getString() );
                            println();
                            dontPrintBlankLine = true;
                            thisIndent = indent;
                        }
                        else if( token.getString().equals( "}" ) )
                        {
                            // if it is followed by a ";", we should not do a new line yet!!!
                            if( printCount > 0 )
                            {
                                println();
                            }
                            print( token.getString() );
                            indent--;
                            thisIndent = indent;
                            println();
                            /*if(indent == 1) // an extra line at level 1 (at the end of a method)
							{
								println();
							}*/
                            dontPrintBlankLine = true;
                        }
                        else
                        {
                            if( useWhitespace( token, previous ) )
                            {
                                print( " " );
                            }
                            print( token.getString() );
                        }
                        
                        if( token.getString().equals( "(" ) )
                        {
                            bracketIndent++;
                        }
                        else if( token.getString().equals( ")" ) )
                        {
                            bracketIndent--;
                        }
                        
                        if( token.getString().equals( ";" ) )
                        {
                            // special case for trailing comment?
                            if( bracketIndent == 0 )
                            {
                                println();
                                dontPrintBlankLine = true;
                            }
                        }
                        break;
                    }
                    
                    default:
                    {
                        System.out.println( "Unknown token type " + token );
                        break;
                    }
                }
                ;
                previous = token;
            }
            
            printResults();
            tokeniser.close();
        }
        catch( IOException e )
        {
            e.printStackTrace();
        }
        return again;
    }
    
    
    /**
     * This function prints out the resulting code
     * The code is already indented, but it still needs the blocks to be re-orded
     * this is done by...
     *
     * printResults
     * @post
     * @pre
     * @throws
     * @returns
     * @param
     */
    void printResults()
    {
        //System.err.println("There are " + codeSections.size() + " code sections");
        
        if( ! foundHeader )
        {
            writeLine( "// $" + "Header" + "$" );
            // this is done in a number of strings so it isn't expanded by CVS.
        }
        
        Vector nonStatic = new Vector();
        
        for( int i = 0; i < codeSections.size(); i++ )
        {
            BCodeSection section = ( BCodeSection ) codeSections.get( i );
            
            //String header = "/* section " + section.getName() + " type " + section.getType() + " scope " + section.getScope() + " */\n";
            //section.prependLine(header);
            
            if( section.getType() == BCodeSection.STATIC )
            {
                if( nonStatic.size() > 0 )
                {
                    writeSections( sortSections( nonStatic ) );
                    nonStatic.clear();
                }
                
                writeSection( section );
            }
            else
            {
                nonStatic.add( section );
            }
        }
        if( nonStatic.size() > 0 )
        {
            writeSections( sortSections( nonStatic ) );
        }
        
        if( printWarning )
        {
            System.err.println( "This file is missing required tags.  Please fill them in" );
            again = true;
        }
        if( warnMissingBrace )
        {
            again = true;
            System.err.println( "Missing braces have been added." );
        }
        
    }
    
    
    /**
     * useWhitespace
     * @post a decision is made on the application of whitespace
     * @pre the ruleset is created
     * @throws none
     * @returns boolean True = use Whitespace; false = don't
     * @param the current token, and previous token or null for first
     */
    
    
    boolean useWhitespace( BToken token, BToken previous )
    {
        if( matchRule( token, previous ) )
        {
            return false;
        }
        else if( printCount == 0 )
        {
            return false;
        }
        return true;
    }
    
    
    /**
     * matchRule
     * @post a decision is made on the number of rules matched
     * @pre the ruleset is initalised
     * @throws none
     * @returns true = at least one match, false = no matches
     * @param the current token, and previous token or null for first
     */
    
    boolean matchRule( BToken token, BToken previous )
    {
        // search through the rules and see if any match
        // if so, return true
        
        // we need a rule set
        // and a rule object...
        if( ruleSet != null )
        {
            for( int i = 0; i < ruleSet.size(); i++ )
            {
                BTokenRule rule = ( BTokenRule ) ruleSet.get( i );
                if( rule.matchRule( token, previous ) )
                {
                    return true;
                }
            }
        }
        
        // if we reach here, then there are no matching rules
        return false;
    }
    
    /**
     * createRuleSet
     * @post the rule set is initailsed
     * @pre none
     * @throws none
     * @returns void
     * @param void
     */
    
    void createRuleSet()
    {
        ruleSet = new Vector();
        
        //                         previous                    token
        //                         token type         string | token type         string
        //ruleSet.add(new BTokenRule(BTokeniser.TT_ANY, null,    BTokeniser.TT_ANY, null));
        ruleSet.add( new BTokenRule( BTokeniser.TT_SYMBOL, "(", BTokeniser.TT_SYMBOL, ")" ) );
        ruleSet.add( new BTokenRule( BTokeniser.TT_SYMBOL, "[", BTokeniser.TT_SYMBOL, "]" ) );
        ruleSet.add( new BTokenRule( BTokeniser.TT_ANY, null, BTokeniser.TT_SYMBOL, ";" ) );
        ruleSet.add( new BTokenRule( BTokeniser.TT_WORD, null, BTokeniser.TT_SYMBOL, "." ) );
        ruleSet.add( new BTokenRule( BTokeniser.TT_SYMBOL, ".", BTokeniser.TT_WORD, null ) );
        ruleSet.add( new BTokenRule( BTokeniser.TT_SYMBOL, ".", BTokeniser.TT_SYMBOL, "*" ) );
        ruleSet.add( new BTokenRule( BTokeniser.TT_ANY, null, BTokeniser.TT_SYMBOL, "," ) );
        ruleSet.add( new BTokenRule( BTokeniser.TT_SYMBOL, ")", BTokeniser.TT_SYMBOL, "." ) );
        ruleSet.add( new BTokenRule( BTokeniser.TT_WORD, null, BTokeniser.TT_SYMBOL, ":" ) );
        ruleSet.add( new BTokenRule( BTokeniser.TT_SYMBOL, "<", BTokeniser.TT_SYMBOL, "<" ) );
        ruleSet.add( new BTokenRule( BTokeniser.TT_SYMBOL, ">", BTokeniser.TT_SYMBOL, ">" ) );
        ruleSet.add( new BTokenRule( BTokeniser.TT_SYMBOL, ">", BTokeniser.TT_SYMBOL, ">=" ) );
        ruleSet.add( new BTokenRule( BTokeniser.TT_SYMBOL, "<", BTokeniser.TT_SYMBOL, "<=" ) );
        ruleSet.add( new BTokenRule( BTokeniser.TT_SYMBOL, "<<", BTokeniser.TT_SYMBOL, "=" ) );
        ruleSet.add( new BTokenRule( BTokeniser.TT_SYMBOL, ">>", BTokeniser.TT_SYMBOL, "=" ) );
        ruleSet.add( new BTokenRule( BTokeniser.TT_WORD, null, BTokeniser.TT_SYMBOL, "++" ) );
        ruleSet.add( new BTokenRule( BTokeniser.TT_WORD, null, BTokeniser.TT_SYMBOL, "--" ) );
        ruleSet.add( new BTokenRule( BTokeniser.TT_SYMBOL, "++", BTokeniser.TT_WORD, null ) );
        ruleSet.add( new BTokenRule( BTokeniser.TT_SYMBOL, "--", BTokeniser.TT_WORD, null ) );
        ruleSet.add( new BTokenRule( BTokeniser.TT_WORD, null, BTokeniser.TT_SYMBOL, "(" ) );
        
        
        //		ruleSet.add(new BTokenRule(BTokeniser.TT_ANY,    null,   BTokeniser.TT_SYMBOL, "("));
        
        
        // these are the basic rules which we don't want to match.. (they were for testing...)
        //ruleSet.add(new BTokenRule(BTokeniser.TT_WORD, null,    BTokeniser.TT_WORD, null));
        //ruleSet.add(new BTokenRule(BTokeniser.TT_WORD, null,    BTokeniser.TT_NUMBER, null));
        //ruleSet.add(new BTokenRule(BTokeniser.TT_NUMBER, null,    BTokeniser.TT_NUMBER, null));
        //ruleSet.add(new BTokenRule(BTokeniser.TT_NUMBER, null,    BTokeniser.TT_WORD, null));
    }
    
    
    /**
     * sortMethods
     * @post the methods are sorted into the correct order
     * @pre
     * @throws
     * @returns a new vector
     * @param the inital vector
     */
    
    private Vector sortMethods( Vector sections )
    {
        // we just need to find all the get & set pairs
        // do
        // is
        // on
        // add+remove in matching pairs
        // and anything left over
        Vector results = new Vector();
        
        Vector constructors = new Vector();
        Vector gets = new Vector();
        Vector sets = new Vector();
        Vector dos = new Vector();
        Vector iss = new Vector();
        Vector ons = new Vector();
        Vector adds = new Vector();
        Vector removes = new Vector();
        Vector others = new Vector();
        
        // go through and move all the results to the appropriate sections...
        for( int i = 0; i < sections.size(); i++ )
        {
            BCodeSection section = ( BCodeSection ) sections.get( i );
            if( section.getName() != null )
            {
                if( section.getName().equals( className ) )
                {
                    constructors.add( section );
                }
                else if( section.getName().startsWith( "get" ) )
                {
                    gets.add( section );
                }
                else if( section.getName().startsWith( "set" ) )
                {
                    sets.add( section );
                }
                else if( section.getName().startsWith( "do" ) )
                {
                    dos.add( section );
                }
                else if( section.getName().startsWith( "is" ) )
                {
                    iss.add( section );
                }
                else if( section.getName().startsWith( "on" ) )
                {
                    ons.add( section );
                }
                else if( section.getName().startsWith( "add" ) )
                {
                    adds.add( section );
                }
                else if( section.getName().startsWith( "remove" ) )
                {
                    removes.add( section );
                }
                else
                {
                    others.add( section );
                }
            }
            else
            {
                others.add( section );
            }
        }
        
        // constructors
        while( constructors.size() > 0 )
        {
            BCodeSection section = ( BCodeSection ) constructors.get( 0 );
            results.add( section );
            constructors.remove( 0 );
        }
        // now, pair up get and sets...
        while( gets.size() > 0 )
        {
            BCodeSection section = ( BCodeSection ) gets.get( 0 );
            results.add( section );
            gets.remove( 0 );
            
            // see if there is a matching set
            String matchName = "set" + section.getName().substring( 3 );
            for( int i = 0; i < sets.size(); i++ )
            {
                BCodeSection section2 = ( BCodeSection ) sets.get( 0 );
                if( section2.getName().equals( matchName ) )
                {
                    results.add( section2 );
                    sets.remove( i );
                    break;
                }
            }
        }
        
        // and remaining sets
        while( sets.size() > 0 )
        {
            BCodeSection section = ( BCodeSection ) sets.get( 0 );
            results.add( section );
            sets.remove( 0 );
        }
        
        // the do's
        while( dos.size() > 0 )
        {
            BCodeSection section = ( BCodeSection ) dos.get( 0 );
            results.add( section );
            dos.remove( 0 );
        }
        
        // the is's
        while( iss.size() > 0 )
        {
            BCodeSection section = ( BCodeSection ) iss.get( 0 );
            results.add( section );
            iss.remove( 0 );
        }
        
        // the on's
        while( ons.size() > 0 )
        {
            BCodeSection section = ( BCodeSection ) ons.get( 0 );
            results.add( section );
            ons.remove( 0 );
        }
        
        // now, pair up add and removes...
        while( adds.size() > 0 )
        {
            BCodeSection section = ( BCodeSection ) adds.get( 0 );
            results.add( section );
            adds.remove( 0 );
            
            // see if there is a matching set
            String matchName = "remove" + section.getName().substring( 3 );
            for( int i = 0; i < removes.size(); i++ )
            {
                BCodeSection section2 = ( BCodeSection ) removes.get( 0 );
                if( section2.getName().equals( matchName ) )
                {
                    results.add( section2 );
                    removes.remove( i );
                    break;
                }
            }
        }
        
        // and remaining removes
        while( removes.size() > 0 )
        {
            BCodeSection section = ( BCodeSection ) removes.get( 0 );
            results.add( section );
            removes.remove( 0 );
        }
        
        // and the left overs...
        while( others.size() > 0 )
        {
            BCodeSection section = ( BCodeSection ) others.get( 0 );
            results.add( section );
            others.remove( 0 );
        }
        return results;
    }
    
    
    /**
     * Please complete the missing tags for sortSections
     * @post the sections are in the correct order
     * @pre
     * @throws none
     * @returns a new vector of the same sections
     * @param the inital vector of sections
     */
    
    private Vector sortSections( Vector sections )
    {
        Vector variables = new Vector();
        Vector methods = new Vector();
        Vector classes = new Vector();
        Vector result = new Vector();
        
        // firsts, split up the block types
        
        for( int i = 0; i < sections.size(); i++ )
        {
            BCodeSection section = ( BCodeSection ) sections.get( i );
            
            switch( section.getType() )
            {
                case BCodeSection.METHOD:
                {
                    methods.add( section );
                    break;
                }
                case BCodeSection.IV:
                {
                    variables.add( section );
                    break;
                }
                case BCodeSection.CLASS:
                {
                    classes.add( section );
                    break;
                }
                default:
                {
                    System.out.println( "sorting of this code block not supported" );
                }
            }
        }
        // Next.... instance variables
        // in what order?
        // the same that they were declared in....?
        
        // public, protected, default, private
        for( int i = 0; i < variables.size(); i++ )
        {
            BCodeSection section = ( BCodeSection ) variables.get( i );
            if( section.getScope() == BCodeSection.PUBLIC )
            {
                result.add( section );
            }
        }
        
        for( int i = 0; i < variables.size(); i++ )
        {
            BCodeSection section = ( BCodeSection ) variables.get( i );
            if( section.getScope() == BCodeSection.PROTECTED )
            {
                result.add( section );
            }
        }
        
        for( int i = 0; i < variables.size(); i++ )
        {
            BCodeSection section = ( BCodeSection ) variables.get( i );
            if( section.getScope() == BCodeSection.DEFAULT )
            {
                result.add( section );
            }
        }
        
        for( int i = 0; i < variables.size(); i++ )
        {
            BCodeSection section = ( BCodeSection ) variables.get( i );
            if( section.getScope() == BCodeSection.PRIVATE )
            {
                result.add( section );
            }
        }
        
        
        
        // next we want methods
        // these are the hard bits.....
        // the order is constructos.. (what is the class name....?)
        // public
        // protected
        // default
        // private
        Vector methodList = new Vector();
        for( int i = 0; i < methods.size(); i++ )
        {
            BCodeSection section = ( BCodeSection ) methods.get( i );
            if( section.getScope() == BCodeSection.PUBLIC )
            {
                methodList.add( section );
            }
        }
        methodList = sortMethods( methodList );
        for( int i = 0; i < methodList.size(); i++ )
        {
            result.add( methodList.get( i ) );
        }
        
        // Private methods
        methodList.clear();
        for( int i = 0; i < methods.size(); i++ )
        {
            BCodeSection section = ( BCodeSection ) methods.get( i );
            if( section.getScope() == BCodeSection.PROTECTED )
            {
                methodList.add( section );
            }
        }
        methodList = sortMethods( methodList );
        for( int i = 0; i < methodList.size(); i++ )
        {
            result.add( methodList.get( i ) );
        }
        
        // default methods
        methodList.clear();
        for( int i = 0; i < methods.size(); i++ )
        {
            BCodeSection section = ( BCodeSection ) methods.get( i );
            if( section.getScope() == BCodeSection.DEFAULT )
            {
                methodList.add( section );
            }
        }
        methodList = sortMethods( methodList );
        for( int i = 0; i < methodList.size(); i++ )
        {
            result.add( methodList.get( i ) );
        }
        
        // private methods
        methodList.clear();
        for( int i = 0; i < methods.size(); i++ )
        {
            BCodeSection section = ( BCodeSection ) methods.get( i );
            if( section.getScope() == BCodeSection.PRIVATE )
            {
                methodList.add( section );
            }
        }
        methodList = sortMethods( methodList );
        for( int i = 0; i < methodList.size(); i++ )
        {
            result.add( methodList.get( i ) );
        }
        
        // classes, just dump them in...
        for( int i = 0; i < classes.size(); i++ )
        {
            BCodeSection section = ( BCodeSection ) classes.get( i );
            result.add( section );
        }
        
        
        // we need to look at all sorts of stuff....
        return result;
    }
    
    
    /**
     * writeSection
     * @post the section is written to the output
     * @pre
     * @throws none
     * @returns void
     * @param a section
     */
    
    private void writeSection( BCodeSection section )
    {
        
        addMissingTags( section );
        
        for( int line = 0; line < section.getCount(); line++ )
        {
            writeLine( section.getLine( line ) );
        }
    }
    
    /**
     * writeSections
     * @post the sections are written to the output
     * @pre none
     * @throws none
     * @returns void
     * @param a vector of sections
     */
    
    private void writeSections( Vector sections )
    {
        for( int i = 0; i < sections.size(); i++ )
        {
            BCodeSection section = ( BCodeSection ) sections.get( i );
            writeSection( section );
        }
    }
    
    
    /**
     * writeLine
     * @post the string is written to the output
     * @pre
     * @throws none
     * @returns void
     * @param a string
     *
     * This writes a line of text to the actual output stream
     * It is called by the printResults related code.
     * It appends a newline character which is appropriate for the platform on which it is running (eventualy)
     */
    
    private void writeLine( String str )
    {
        try
        {
            str = str + "\n";
            //System.getProperty ( "line.separator" );
            outStream.write( str.getBytes() );
        }
        catch( IOException e )
        {
            System.err.println( "Error writing output" );
        }
    }
}

